#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlwriter.h>
#include "rss.h"


/*
Introduction

libxml2 is a library of functions for handling XML data.


A simple example:

#include <stdio.h>
#include <libxml/tree.h>


int
main ()
{
  xmlDocPtr doc;
  xmlNodePtr nodeLevel1;
  xmlNodePtr nodeLevel2;

  doc = xmlParseFile ("xmlfile.xml");
  for (nodeLevel1 = doc->children;
       nodeLevel1 != NULL; nodeLevel1 = nodeLevel1->next)
    {
      printf ("%s\n", nodeLevel1->name);
      for (nodeLevel2 = nodeLevel1->children;
           nodeLevel2 != NULL; nodeLevel2 = nodeLevel2->next)
        {
          printf ("\t%s\n", nodeLevel2->name);
        }
    }
  xmlSaveFile ("xmlfile_copy.xml", doc);
  xmlFreeDoc (doc);
  return 0;
}


The above code, compiled with -lxml2, should print out the names of the
elements in the first two elements' depth of an XML file, and save a copy
of the file.


Explanation of Introduction

The xmlDocPtr is a pointer to an xmlDoc structure.
It represents an XML data source.

You load an XML file with the xmlParseFile() function, which takes
as a parameter the name of an XML file and returns a pointer to a
new xmlDoc structure (or NULL on failure). When done, you release
this memory with the xmlFreeDoc() function. You can export an xmlDocPtr's
data as an XML file with the xmlSaveFile() function.



The xmlNodePtr points to a single element or node of an XML document.
Each xmlNode has a .children member which is an xmlNodePtr to the first
of this node's children. Each xmlNode has a .name member which is a
string containing the name of the element it represents, or the word "text"
for a text node.

The xmlNodePtr is the basic structure used to traverse an XML document
with libxml2. It contains several xmlNodePtrs which can be used to move
around the document. If there is no other node in a particular direction,
the pointer is NULL.

xmlNodePtr->children	The first child of the node
xmlNodePtr->last	The node's last child
xmlNodePtr->parent	The current node's parent node

xmlNodePtr->next	The next sibling node
xmlNodePtr->prev	The previous sibling node

xmlNodePtr->doc		The xmlDocPtr for the document containing this node


<node>
        <node> ->parent
                <node>
                        <node> ->prev
                        </node>
                        <node>
                        </node>
                </node>
                <node> You Are Here
                        <node> ->children
                        </node>
                        <node>
                        </node>
                        <node> ->last
                        </node>
                </node>
                <node> ->next
                        <node>
                        </node>
                </node>
        </node>
</node>



Although the above diagram suggests that nodes are simply elements,
be aware that areas of whitespace between elements are also nodes.
This makes the above example faulty if it is taken literally, because
the child and neighbour node pointers would point to the whitespace
between elements.
*/


#ifndef _WIN32
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif
#ifndef BAD_CAST
#define BAD_CAST (const xmlChar *)
#endif


#define RSS_V0_91_S "RSS v0.91"
#define RSS_V0_92_S "RSS v0.92"
#define RSS_V0_93_S "RSS v0.93"
#define RSS_V0_94_S "RSS v0.94"
#define RSS_V1_0_S  "RSS v1.0"
#define RSS_V2_0_S  "RSS v2.0"
#define ATOM_V0_1_S "ATOM v0.1"
#define ATOM_V0_2_S "ATOM v0.2"
#define ATOM_V0_3_S "ATOM v0.3"


#ifdef  DEBUG
void
rss_st_rss_t_sanity_check (st_rss_t *rss)
{
  int i = 0;

  for (; i < rss->item_count; i++)
    printf ("pos:%d\n"
            "title: %s\n"
            "url: %s\n"
            "date: %ld\n"
            "desc: %s\n\n",
      i,
      rss->item[i].title,
      rss->item[i].url,
      rss->item[i].date,
      rss->item[i].desc);

  printf ("rss->item_count: %d\n\n", rss->item_count);
}
#endif


char *
rss_utf8_enc (const char *in, const char *encoding)
{
  static xmlChar out[RSSMAXBUFSIZE * 2];
  int temp = 0;
  int size = 0;
  xmlCharEncodingHandlerPtr handler;

  if (!in)
    return NULL;

  handler = xmlFindCharEncodingHandler (encoding);

  if (!handler)
    return NULL;

  temp = strlen (in);
  size = sizeof (out);

  handler->input (out, &size, BAD_CAST in, &temp);

  return (char *) out;
}


typedef struct
{
  int version;
  const char *version_s;

  const char *magic_s;
} st_rss_version_t;


static st_rss_version_t rss_version[] = {
  {RSS_V0_91, RSS_V0_91_S, "0.91"},
  {RSS_V0_92, RSS_V0_92_S, "0.92"},
  {RSS_V0_93, RSS_V0_93_S, "0.93"},
  {RSS_V0_94, RSS_V0_94_S, "0.94"},
  {RSS_V1_0,  RSS_V1_0_S,  "1.0"},
  {RSS_V2_0,  RSS_V2_0_S,  "2.0"},
  {RSS_V2_0,  RSS_V2_0_S,  "2"},
  {ATOM_V0_1, ATOM_V0_1_S, "0.1"},
  {ATOM_V0_2, ATOM_V0_2_S, "0.2"},
  {ATOM_V0_3, ATOM_V0_3_S, "0.3"},
  {0, NULL, NULL}
};


const char *
rss_get_version_s_by_magic (const char *m)
{
  int i = 0;

  for (; rss_version[i].version; i++)
    if (!strcmp (rss_version[i].magic_s, m))
      return rss_version[i].version_s;
  return NULL;
}


const char *
rss_get_version_s_by_id (int version)
{
  int i = 0;

  for (; rss_version[i].version; i++)
    if (rss_version[i].version == version)
      return rss_version[i].version_s;
  return NULL;
}

const char *
rss_get_version_s (st_rss_t * rss)
{
  return rss_get_version_s_by_id (rss->version);
}


unsigned int
rss_item_count (st_rss_t * rss)
{
  return rss->item_count;
}


st_rss_item_t *
rss_get_item (st_rss_t * rss, unsigned int n)
{
  return &rss->item[n];
}


#if 0
void
rsstool_set_version (const char *version)
{
  const st_rss_version_t *v = NULL;

  if (!(*version))
    fprintf (stderr, "WARNING: no version in %d\n", rsstool.article_count);

  v = rsstool_get_rss_version_by_magic (version);
  if (v)
    rsstool.rss[rsstool.article_count].version = v->version;

#ifdef  DEBUG
  printf ("link[%d]: %s\n", rsstool.article_count, rsstool.rss[rsstool.article_count].version);
  fflush (stdout);
#endif
}


void
rsstool_set_site (const char *site)
{
  if (!(*site))
    fprintf (stderr, "WARNING: no site discovered\n");

  strncpy (rsstool.rss[rsstool.article_count].site, site, RSSMAXBUFSIZE)[RSSMAXBUFSIZE - 1] = 0;

#ifdef  DEBUG
  printf ("site[%d]: %s\n", rsstool.article_count, rsstool.rss[rsstool.article_count].site);
  fflush (stdout);
#endif
}
#endif


time_t
parse_date (const char *s)
{
  int i = 0;
  char y[100], m[100], d[100];
  char h[100], min[100];
//  char sec[100];
  struct tm time_tag;
  time_t t = time (0);
  const char *month_s[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", NULL};

  *y = *m = *d = *h = *min = 0;

  if (s[10] == 'T')                     // YYYY-MM-DDT00:00+00:00
    {
      sscanf (s, " %4s-%2s-%2sT%2s:%2s", y, m, d, h, min);
    }
  else if (s[3] == ',' && s[4] == ' ')  // Mon, 31 Jul 2006 15:05:00 GMT
    {
      sscanf (s + 5, "%2s %s %4s %2s:%2s", d, m, y, h, min);

      for (i = 0; month_s[i]; i++)
        if (!stricmp (m, month_s[i]))
          {
            sprintf (m, "%d", i + 1);
            break;
          }
    }
  else if (s[4] == '-' && s[7] == '-')  // 2006-07-19
    {
      sscanf (s, "%4s-%2s-%2s", y, m, d);
    }
  else                                  // YYYYMMDDTHHMMSS
    {
//      sscanf (s, " %4s%2s%2sT", y, m, d);
    }

  memset (&time_tag, 0, sizeof (struct tm));

  if (*y)
    time_tag.tm_year = strtol (y, NULL, 10) - 1900;
  if (*m)
    time_tag.tm_mon = strtol (m, NULL, 10) - 1;
  if (*d)
    time_tag.tm_mday = strtol (d, NULL, 10);
  if (*h)
    time_tag.tm_hour = strtol (h, NULL, 10);
  if (*min)
    time_tag.tm_min = strtol (min, NULL, 10);

  t = mktime (&time_tag);

  return t;
}


int
rss_demux (const char *fname)
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  int version = -1;
  char *p = NULL;

  if (!(doc = xmlParseFile (fname)))
    {
      fprintf (stderr, "ERROR: cannot read %s\n", fname);
      return -1;
    }

  cur = xmlDocGetRootElement (doc);

  if (!cur)
    return -1;

  if (!cur->name)
    return -1;

#ifdef  DEBUG
  printf ("%s\n", cur->name);
  fflush (stdout);
#endif

  if (!xmlStrcmp (cur->name, BAD_CAST "html")) // not xml
    return -1;

  if (!xmlStrcmp (cur->name, BAD_CAST "feed")) // atom
    {
      if (!xmlHasProp (cur, BAD_CAST "version"))
        return -1;

      if (!(p = (char *) xmlGetProp (cur, BAD_CAST "version")))
        return -1;

      if (!strcmp (p, "0.3"))
        version = ATOM_V0_3;
      else if (!strcmp (p, "0.2"))
        version = ATOM_V0_2;
      else if (!strcmp (p, "0.1"))
        version = ATOM_V0_1;

      return version;
    }
  else if (!xmlStrcmp (cur->name, BAD_CAST "rss")) // rss
    {
      if (!xmlHasProp (cur, BAD_CAST "version"))
        return -1;

      if (!(p = (char *) xmlGetProp (cur, BAD_CAST "version")))
        return -1;

      if (!strcmp (p, "0.91"))
        version = RSS_V0_91;
      else if (!strcmp (p, "0.92"))
        version = RSS_V0_92;
      else if (!strcmp (p, "0.93"))
        version = RSS_V0_93;
      else if (!strcmp (p, "0.94"))
        version = RSS_V0_94;
      else if (!strcmp (p, "2") || !strcmp (p, "2.0") || !strcmp (p, "2.00"))
        version = RSS_V2_0;

      return version;
    }
  else if (!xmlStrcmp (cur->name, BAD_CAST "rdf") ||
           !xmlStrcmp (cur->name, BAD_CAST "RDF")) // rdf
    {
#if 0
      if (!xmlHasProp (cur, "xmlns"))
        return -1;

      if (!(p = xmlGetProp (cur, "xmlns")))
        return -1;

      // hack
      if (!strcmp (p, "http://my.netscape.com/rdf/simple/0.9/"))
        version = RSS_V0_90;
      else if (!strcmp (p, "http://purl.org/rss/1.0/"))
        version = RSS_V1_0;
#else
      version = RSS_V1_0;
#endif

      return version;
    }

  return -1;
}


static void
rss_read_copy (char *d, xmlDocPtr doc, xmlNodePtr n)
{
  const char *p = (const char *) xmlNodeListGetString (doc, n, 1);

  if (p)
    strncpy (d, p, RSSMAXBUFSIZE)[RSSMAXBUFSIZE-1] = 0;
  else
    *d = 0;
}



st_rss_t *
rss_open (const char *fname)
{
  static st_rss_t rss;
  xmlDocPtr doc;
  xmlNodePtr cur;
  int rdf = 0;

  memset (&rss, 0, sizeof (st_rss_t));

  strncpy (rss.url, fname, RSSMAXBUFSIZE)[RSSMAXBUFSIZE - 1] = 0;
  rss.item_count = 0;

  doc = xmlParseFile (rss.url);
  if (!doc)
    {
      fprintf (stderr, "ERROR: cannot read %s\n", rss.url);
      return NULL;
    }

  rss.version = rss_demux (fname);
#ifdef  DEBUG
  fprintf (stderr, "version: %s\n", rss_get_version_s_by_id (rss.version));
  fflush (stderr);
#endif
  if (rss.version == -1)
    {
      fprintf (stderr, "ERROR: unknown feed format %s\n", rss.url);
      return NULL;
    }

  cur = xmlDocGetRootElement (doc);
  if (!cur)
    {
      fprintf (stderr, "ERROR: empty document %s\n", rss.url);
      xmlFreeDoc (doc);
      return NULL;
    }

  // rdf?
  // TODO: move this to rss_demux()
  if (xmlStrcmp (cur->name, BAD_CAST "rss") != 0 &&
      (!xmlStrcmp (cur->name, BAD_CAST "rdf") ||
       !xmlStrcmp (cur->name, BAD_CAST "RDF")))
    rdf = 1;

  cur = cur->xmlChildrenNode;
  while (cur && xmlIsBlankNode (cur))
    cur = cur->next;
  if (!cur)
    {
//      fprintf (stderr, "");
      return NULL;
    }

  if (xmlStrcmp (cur->name, BAD_CAST "channel"))
    {
      fprintf (stderr, "ERROR: bad document: did not immediately find the rss element\n");
      return NULL;
    }

  if (!rdf) // document is RSS
    cur = cur->xmlChildrenNode;

  while (cur)
    {
      if (!xmlStrcmp (cur->name, BAD_CAST "title"))
        rss_read_copy (rss.title, doc, cur->xmlChildrenNode);
      else if (!xmlStrcmp (cur->name, BAD_CAST "description"))
        rss_read_copy (rss.desc, doc, cur->xmlChildrenNode);
//      else if (!xmlStrcmp (cur->name, BAD_CAST "link"))
//        rss_read_copy (rss.url, doc, cur->xmlChildrenNode);
      else if (!xmlStrcmp (cur->name, BAD_CAST "date") ||
               !xmlStrcmp (cur->name, BAD_CAST "pubDate") ||
               !xmlStrcmp (cur->name, BAD_CAST "dc:date"))
        rss.date = parse_date ((const char *) xmlNodeListGetString (doc, cur->xmlChildrenNode, 1));
      else if (!xmlStrcmp (cur->name, BAD_CAST "channel") && rdf)
        {
          xmlNodePtr pcur = cur->xmlChildrenNode;

          while (pcur)
            {
              if (!xmlStrcmp (pcur->name, BAD_CAST "title"))
                rss_read_copy (rss.title, doc, pcur->xmlChildrenNode);
              else if (!xmlStrcmp (pcur->name, BAD_CAST "description"))
                rss_read_copy (rss.desc, doc, pcur->xmlChildrenNode);
              else if (!xmlStrcmp (pcur->name, BAD_CAST "date") ||
                       !xmlStrcmp (pcur->name, BAD_CAST "pubDate") ||
                       !xmlStrcmp (pcur->name, BAD_CAST "dc:date"))
                rss.date = parse_date ((const char *) xmlNodeListGetString (doc, pcur->xmlChildrenNode, 1));

              pcur = pcur->next;
            }

        }
      else if ((!xmlStrcmp (cur->name, BAD_CAST "item")))
        {
          xmlNodePtr pcur = cur->xmlChildrenNode;
          st_rss_item_t *item = &rss.item[rss.item_count];
          int found = 0;

          while (pcur)
            {
#ifdef  DEBUG
              printf ("%s\n", pcur->name);
              fflush (stdout);
#endif

              if (!xmlStrcmp (pcur->name, BAD_CAST "title"))
                {
                  rss_read_copy (item->title, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xmlStrcmp (pcur->name, BAD_CAST "link") ||
                       !xmlStrcmp (pcur->name, BAD_CAST "guid"))
                {
                  rss_read_copy (item->url, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xmlStrcmp (pcur->name, BAD_CAST "description"))
                {
                  rss_read_copy (item->desc, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xmlStrcmp (pcur->name, BAD_CAST "date") ||
                       !xmlStrcmp (pcur->name, BAD_CAST "pubDate") ||
                       !xmlStrcmp (pcur->name, BAD_CAST "dc:date"))
                { 
                  item->date = parse_date ((const char *) xmlNodeListGetString (doc, pcur->xmlChildrenNode, 1));
                  found = 1;
                }
              else if (!xmlStrcmp (pcur->name, BAD_CAST "text"))
                {
                }
              else
                {
                  if (!found) // possibly malformed feed
                    break;
                  else
                    found = 0;
                }

              pcur = pcur->next;
            }

          rss.item_count++;

          if (rss.item_count == RSSMAXITEM)
            break;
        }

//      rss.item_count++;

      cur = cur->next;
    }

#ifdef  DEBUG
  rss_st_rss_t_sanity_check (&rss);
  fflush (stdout);
#endif

  return &rss;
}


int
rss_close (st_rss_t *rss)
{
  (void) rss;
  return 0;
}


int
rss_write (FILE *fp, st_rss_t *rss, int version)
{
#define XMLPRINTF(s) xmlTextWriterWriteString(writer,BAD_CAST s)
  xmlTextWriterPtr writer;
  xmlBufferPtr buffer;
  char buf[RSSMAXBUFSIZE];
  unsigned int i = 0;

  if (!fp)
    return -1;

  if (!rss)
    return -1;

  if (!(buffer = xmlBufferCreate ()))
    return -1;

  if (!(writer = xmlNewTextWriterMemory (buffer, 0)))
    return -1;

  if (version != 1) // default to RSS 2.0
    version = 2;

  xmlTextWriterStartDocument (writer, NULL, "UTF-8", NULL);

//  xmlTextWriterWriteComment (writer, "comment");

  if (version == 2)
    {
      xmlTextWriterStartElement (writer, BAD_CAST "rss");  // <rss>

      xmlTextWriterWriteAttribute (writer, BAD_CAST "version", BAD_CAST "2.0");
    }
  else
    {
      xmlTextWriterStartElement (writer, BAD_CAST "rdf:RDF");  // <rdf:RDF>

      xmlTextWriterWriteAttribute (writer, BAD_CAST "xmlns", BAD_CAST "http://purl.org/rss/1.0/"); // specs?
    }

  XMLPRINTF("\n  ");

  xmlTextWriterStartElement (writer, BAD_CAST "channel"); // <channel>

  XMLPRINTF("\n    ");

  xmlTextWriterWriteElement (writer, BAD_CAST "title", BAD_CAST rss->title);

  XMLPRINTF("\n    ");

  xmlTextWriterWriteElement (writer, BAD_CAST "link", BAD_CAST rss->url);

  XMLPRINTF("\n    ");

  xmlTextWriterWriteElement (writer, BAD_CAST "description", BAD_CAST rss->desc);

#if 0
  XMLPRINTF("\n    ");

  xmlTextWriterWriteFormatElement (writer, BAD_CAST "dc:date", "%ld", BAD_CAST rss->date);
#endif

  if (version == 1)
    {
      XMLPRINTF("\n    ");

      xmlTextWriterStartElement (writer, BAD_CAST "items"); // <items>

      XMLPRINTF("\n      ");

      xmlTextWriterStartElement (writer, BAD_CAST "rdf:Seq"); // <rdf:Seq>

      for (i = 0; i < rss_item_count (rss); i++)
        {
          sprintf (buf, "\n        <rdf:li rdf:resource=\"%s\"/>", rss->item[i].url);
          XMLPRINTF(buf);
        }

      XMLPRINTF("\n      ");

      xmlTextWriterEndElement (writer); // </rdf:Seq>

      XMLPRINTF("\n    ");

      xmlTextWriterEndElement (writer); // </items>

      XMLPRINTF("\n  ");

      xmlTextWriterEndElement (writer); // </channel>
    }

  for (i = 0; i < rss_item_count (rss); i++)
    {
      XMLPRINTF("\n    ");

      xmlTextWriterStartElement (writer, BAD_CAST "item"); // <item>

      if (version == 1)
        xmlTextWriterWriteAttribute (writer, BAD_CAST "rdf:about", BAD_CAST rss->item[i].url);

      XMLPRINTF("\n      ");

      xmlTextWriterWriteElement (writer, BAD_CAST "title", BAD_CAST rss->item[i].title);

      XMLPRINTF("\n      ");

      xmlTextWriterWriteElement (writer, BAD_CAST "link", BAD_CAST rss->item[i].url);

      XMLPRINTF("\n      ");

      xmlTextWriterWriteElement (writer, BAD_CAST "description", BAD_CAST rss->item[i].desc);

      XMLPRINTF("\n      ");

      xmlTextWriterWriteFormatElement (writer, BAD_CAST "dc:date", "%ld", rss->item[i].date);

      XMLPRINTF("\n    ");

      xmlTextWriterEndElement (writer); // </item>
    }

  if (version == 2)
    {
      XMLPRINTF("\n  ");

      xmlTextWriterEndElement (writer); // </channel>
    } 

  XMLPRINTF("\n");

  xmlTextWriterEndDocument (writer);  // </rss> or </rdf>

  xmlFreeTextWriter (writer);

  fputs ((const char *) buffer->content, fp);

  xmlBufferFree (buffer);

  return 0;
}
